;;; find-cmd-widget.el --- Build a valid find(1) command with widgets -*- lexical-binding: t -*-

;; Copyright (C) 2019 Free Software Foundation, Inc

;; Author: Michael Heerdegen <michael_heerdegen@web.de>
;; Maintainer: Michael Heerdegen <michael_heerdegen@web.de>
;; Created: 10 Jun 2019
;; Keywords: convenience
;; Compatibility: GNU Emacs 26
;; Version: 0.1
;; Package-Requires: ((emacs "26"))


;; This file is not part of GNU Emacs.

;; GNU Emacs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; ...

;;; Code:

;;;; Requirements

(require 'widget)
(require 'find-cmd)
(require 'subr-x)
(require 'cus-edit)
(eval-when-compile (require 'wid-edit))


;;;; Variables

(defvar find-cmd-widget-find-arg-alist
  ;; elements are (symbol make-widget-fun category annotation hint-string)
  ;;               0      1               2        3          4
  '(
    ;; generic tests
    (amin       find-cmd-widget-make-time-minutes-widget   "Tests"
                "Minutes since file last accessed")
    (anewer     find-cmd-widget-make-xnewer-widget         "Tests"
                "Last accessed more recently than a specified file was modified")
    (atime      find-cmd-widget-make-time-days-widget      "Tests"
                "Days since file last accessed (rounded down; see also `daystart`)")
    (cmin       find-cmd-widget-make-time-minutes-widget   "Tests"
                "Minutes since last status change")
    (cnewer     find-cmd-widget-make-xnewer-widget         "Tests"
                "Status changed more recently than a specified file was modified")
    (ctime      find-cmd-widget-make-time-days-widget      "Tests"
                "Days since last status change (rounded down; see also `daystart`)")
    (empty      find-cmd-widget-make-choice-widget         "Tests"
                "Empty regular file or directory")
    (executable find-cmd-widget-make-choice-widget         "Tests"
                "File is executable/ Dir is searchable")
    (false      find-cmd-widget-make-choice-widget         "Tests"
                "Always false")
    (fstype     find-cmd-widget-make-editable-field-widget "Tests")
    (gid        find-cmd-widget-make-editable-field-widget "Tests")
    (group      find-cmd-widget-make-editable-field-widget "Tests")
    (ilname     find-cmd-widget-make-lname-widget          "Tests"
                "Case insensitive version of `lname`"
                "shell glob pattern; must match the complete name")
    (iname      find-cmd-widget-make-editable-field-widget "Tests"
                "Like `name` but match is case insensitive"
                "shell glob pattern; must match the complete name")
    (inum       find-cmd-widget-make-editable-field-widget "Tests")
    (ipath      find-cmd-widget-make-editable-field-widget "Tests"
                "Like `path` but match is case insensitive"
                "shell glob pattern; must match the complete path")
    (iregex     find-cmd-widget-make-editable-field-widget "Tests")
    ;; (iwholename (1) "tests") ; same as `ipath`
    (links      find-cmd-widget-make-links-widget          "Tests"
                "Number of hard links")
    (lname      find-cmd-widget-make-lname-widget          "Tests"
                "symlink whose contents match a shell pattern"
                "shell glob pattern; must match the complete name")
    (mmin       find-cmd-widget-make-time-minutes-widget   "Tests"
                "Minutes since last data change")
    (mtime      find-cmd-widget-make-time-days-widget      "Tests"
                "Days since last data change (rounded down; see also `daystart`)")
    (name       find-cmd-widget-make-editable-field-widget "Tests"
                "Match shell pattern against base of file name"
                "shell glob pattern; must match the complete name")
    (newer      find-cmd-widget-make-editable-field-widget "Tests")
    (nogroup    find-cmd-widget-make-choice-widget         "Tests"
                "No group corresponds to file's numeric group ID")
    (nouser     find-cmd-widget-make-choice-widget         "Tests"
                "No user corresponds to file's numeric user ID")
    (path       find-cmd-widget-make-editable-field-widget "Tests"
                "Match shell pattern against complete file name"
                "shell glob pattern; must match the complete path")
    (perm       find-cmd-widget-make-choice-widget         "Tests")
    (readable   find-cmd-widget-make-choice-widget         "Tests" "Readable file")
    (regex      find-cmd-widget-make-editable-field-widget "Tests")
    (samefile   find-cmd-widget-make-editable-field-widget "Tests")
    (size       find-cmd-widget-make-editable-field-widget "Tests")
    (true       find-cmd-widget-make-choice-widget         "Tests" "Always true")
    (type       find-cmd-widget-make-type-widget           "Tests" "File Type")
    (uid        find-cmd-widget-make-editable-field-widget "Tests")
    (used       find-cmd-widget-make-editable-field-widget "Tests")
    (user       find-cmd-widget-make-editable-field-widget "Tests"
                "File owned by specified user" "numeric user ID allowed")
    ;; (wholename  (1) "tests") ; same as `path`
    (writable   find-cmd-widget-make-choice-widget         "Tests" "Writable file")
    (xtype      find-cmd-widget-make-type-widget           "Tests" "File or link target type")

    ;; normal options (always true)
    (daystart              find-cmd-widget-make-choice-widget         "Options")
    (depth                 find-cmd-widget-make-choice-widget         "Options")
    (maxdepth              find-cmd-widget-make-editable-field-widget "Options")
    (mindepth              find-cmd-widget-make-editable-field-widget "Options")
    (mount                 find-cmd-widget-make-choice-widget         "Options")
    (noleaf                find-cmd-widget-make-choice-widget         "Options")
    (ignore_readdir_race   find-cmd-widget-make-choice-widget         "Options")
    (noignore_readdir_race find-cmd-widget-make-choice-widget         "Options")
    (regextype             find-cmd-widget-make-editable-field-widget "Options")
    (xdev                  find-cmd-widget-make-choice-widget         "Options")

    (and   find-cmd-widget-make-combiner-widget "Combiners")
    (not   find-cmd-widget-make-combiner-widget "Combiners")
    (or    find-cmd-widget-make-combiner-widget "Combiners")
    (prune find-cmd-widget-make-combiner-widget "Combiners")

    ;; actions
    (delete  find-cmd-widget-make-choice-widget         "Actions")
    (print0  find-cmd-widget-make-choice-widget         "Actions")
    (printf  find-cmd-widget-make-editable-field-widget "Actions")
    (fprintf find-cmd-widget-make-editable-field-widget "Actions") ;fixme: gets two args!
    (print   find-cmd-widget-make-choice-widget         "Actions")
    (fprint0 find-cmd-widget-make-editable-field-widget "Actions")
    (fprint  find-cmd-widget-make-editable-field-widget "Actions")
    (ls      find-cmd-widget-make-choice-widget         "Actions")
    (fls     find-cmd-widget-make-editable-field-widget "Actions")
    (prune   find-cmd-widget-make-choice-widget         "Actions")
    (quit    find-cmd-widget-make-choice-widget         "Actions")

    ;; these need to be terminated with a ;
    (exec    find-cmd-widget-make-editable-field-widget "Actions")
    (ok      find-cmd-widget-make-editable-field-widget "Actions")
    (execdir find-cmd-widget-make-editable-field-widget "Actions")
    (okdir   find-cmd-widget-make-editable-field-widget "Actions"))
  "Doc...")

(defvar find-cmd-widget-find-global-option-arglist
  '(;; switches
    (L find-cmd-widget-make-choice-widget "Switches")
    (P find-cmd-widget-make-choice-widget "Switches")
    (H find-cmd-widget-make-choice-widget "Switches"))
  "Doc...")

(defvar-local find-cmd-widget-main-widget nil)
(defvar-local find-cmd-widget-preview-widget-1 nil)
(defvar-local find-cmd-widget-preview-widget-2 nil)
(defvar-local find-cmd-widget-warning-widget nil)

(defvar-local find-cmd-widget-current-global-option nil)
(defvar-local find-cmd-widget-warnings '())


;;;; Helpers

(defun find-cmd-widget--just-true (_) t)

(defun find-cmd-widget--warn-arg-not-implemented (arg)
  (when arg (warn "Argument not implemented")))

(defun find-cmd-widget--parse-widget-sexp (sexp)
  `(,@(when-let ((opt (car sexp))) `(,opt)) ,@(cadr sexp)))

(defun find-cmd-widget-create-find-cmd (sexp)
  `(find-cmd ,(find-cmd-widget--parse-widget-sexp (cdr sexp))))

(defun find-cmd-widget-find-call-get-dir+exps (sexp)
  (list (car sexp)
        (mapconcat #'find-to-string (find-cmd-widget--parse-widget-sexp (cdr sexp)) "")))

(defun find-cmd-widget--build-dirs-string (dirs &optional no-shell-quote)
  (mapconcat (if no-shell-quote #'identity #'shell-quote-argument)
             (mapcar #'expand-file-name dirs) " "))

(defun find-cmd-widget-create-find-call (sexp)
  ;; We do that manually because `find-cmd' doesn't yet handle multiple
  ;; search paths
  (let* ((dirs+exps (find-cmd-widget-find-call-get-dir+exps sexp))
         (dirs (car dirs+exps)))
    (concat find-program
            (if dirs
                (concat " " (concat (find-cmd-widget--build-dirs-string dirs)))
              "")
            " "
            (cadr dirs+exps))))

(defun find-cmd-widget-menu-choice-action (&rest args)
  (cl-letf (((symbol-function #'widget-choose)
             #'find-cmd-widget--widget-choose))
    ;; this doesn't notify the parent widget, but it should
    (apply #'widget-choice-action args)))

(defun find-cmd-widget--widget-choose (title items &optional event)
  (let ((widget-menu-max-size 100))
    (if event
        ;; mouse click
        (let* (chosen-item
               (make-item
                (lambda (i)
                  (let* ((s (car i))
                         (e (assoc (intern s) find-cmd-widget-find-arg-alist)))
                    (vector s (lambda () (interactive) (setq chosen-item (cdr i)))
                            :help (nth 3 e))))))
          (popup-menu
           `(,title
             (keymap
              ("By category"
               ,@(mapcar
                  (pcase-lambda (`(,cat . ,items))
                    `(,cat ,@(mapcar make-item items)))
                  (seq-group-by
                   (lambda (item)
                     (nth 2 (assoc (intern (car item)) find-cmd-widget-find-arg-alist)))
                   items)))
              ,@(mapcar make-item (cl-sort items #'string< :key #'car))))
           event)
          chosen-item)
      (setq items (cl-remove-if 'stringp items))
      (let* ((max-item-length (apply #'max
                                     (mapcar (lambda (e) (length (symbol-name (car e))))
                                             find-cmd-widget-find-arg-alist)))
             (max-cat-length  (apply #'max
                                     (mapcar #'length (mapcar #'caddr find-cmd-widget-find-arg-alist))))
             (completions-format 'vertical)
             (val (completing-read
                   (concat title ": ")
                   (lambda (string pred action)
                     (if (eq action 'metadata)
	                 `(metadata
                           (display-sort-function . ,#'identity)
                           (annotation-function
                            .
                            ,(lambda (s)
                               (when-let ((entry (assoc (intern-soft s)
                                                        find-cmd-widget-find-arg-alist)))
                                 (let ((category (nth 2 entry))
                                       (description (nth 3 entry)))
                                   (and (or category description)
                                        (concat (make-string (- max-item-length (length s)) ?\ )
                                                (format "[%s]" category)
                                                (make-string
                                                 (- max-cat-length (length category)) ?\ )
                                                (if description description ""))))))))
                       (complete-with-action action items string pred)))
                   nil t)))
	(if (stringp val)
	    (let ((try (try-completion val items)))
	      (when (stringp try)
		(setq val try))
	      (cdr (assoc val items))))))))

(defun find-cmd-widget--value-get-1 (command)
  (lambda (w) (list command (widget-field-value-get w))))

(defalias 'find-cmd-widget--get-arg
  (pcase-lambda ((and `(,name ,(and (pred functionp) make-widget) . ,rest)
                      (let hint (nth 2 rest))))
    (ignore rest)
    (funcall make-widget name hint)))

(defun find-cmd-widget-find-action (&rest _)
  (interactive)
  (async-shell-command (widget-value find-cmd-widget-preview-widget-2)))

(defun find-cmd-widget-find-dired-action (&rest _)
  (interactive)
  (let ((args (find-cmd-widget-find-call-get-dir+exps (widget-value find-cmd-widget-main-widget))))
    (when (cdr (car args))
      (user-error "find-dired currently supports only one path, sorry"))
    (setf (car args) (if (car args) (find-cmd-widget--build-dirs-string (car args) 'no-quote) "."))
    (apply #'find-dired args)))


;;;; Widget definitions

(define-widget 'find-cmd-preview 'item
  "Doc..."
  :format "%v"
  :value-face 'highlight
  :value-create #'widget-field-value-create
  ;; (lambda (w)
  ;;   (insert (propertize (widget-get w :value) 'face 'success)))
  )

(defvar find-cmd-widget-sign-args
  (mapcar (pcase-lambda (`(,sign ,descr ,val))
            (cons sign
                  `(choice-item :tag ,descr
                                :value-inline ,(lambda (_w) `(,val)))))
          '((< "less than" "-") (> "greater than" "+") (= "exactly" ""))))

(define-widget 'find-cmd-widget-sign 'menu-choice
  "Doc..."
  :format "sign: %v"
  :args (mapcar #'cdr find-cmd-widget-sign-args))

(defun find-cmd-widget-make-choice-widget (name hint)
  (find-cmd-widget--warn-arg-not-implemented hint)
  `(choice-item :tag ,(symbol-name name)
                :value-get ,(lambda (_w) (list name))))

(define-widget 'find-cmd-widget-editable-field 'editable-field
  "Doc..."
  :size 2
  :validate #'find-cmd-widget--just-true
  :value-get (lambda (w) (list (intern (widget-get w :menu-tag)) (widget-field-value-get w)))
  :keymap widget-field-keymap)

(defun find-cmd-widget-make-editable-field-widget (name &optional hint)
  (let ((name-string (symbol-name name)))
    `(find-cmd-widget-editable-field
      :menu-tag ,name-string
      :format ,(concat name-string ": %v" (and hint (format "  (%s)" hint)) "\n"))))

(defun find-cmd-widget-make-lname-widget (name &optional hint)
  (let ((name-string (symbol-name name)))
    `(find-cmd-widget-editable-field
      :menu-tag ,name-string
      :format ,(concat name-string ": %v" (and hint (format "  (%s)" hint)) "\n")
      :value-get
      ,(lambda (w)
         (when (eq 'L (car find-cmd-widget-current-global-option))
           (push (format "Global option \"-L\" and \"-%s\" used together" name-string)
                 find-cmd-widget-warnings))
         (list name (widget-field-value-get w))))))

(define-widget 'find-cmd-widget-n 'find-cmd-widget-editable-field
  "Widget to input find(1) n option arguments.")

(defun find-cmd-widget-make-n-widget (tag &optional hint)
  `(find-cmd-widget-n
    :menu-tag ,tag
    :format ,(concat tag ": %v" (and hint (format "  (%s)" hint)) "\n")))

(defun find-cmd-widget-make-signed-n-widget (name hint unit &optional default-sign)
  (find-cmd-widget--warn-arg-not-implemented hint)
  (let ((sname (symbol-name name)))
    `(group
      :tag ,sname
      :format ,(format "%s          [%s]\n%%v"
                       sname
                       (nth 3 (assoc name find-cmd-widget-find-arg-alist)))
      :value-get ,(lambda (w) `(,name ,(apply #'concat (widget-editable-list-value-get w))))
      :args
      ((find-cmd-widget-sign
        ,@(and default-sign `(:explicit-choice ,(assoc default-sign find-cmd-widget-sign-args))))
       ,(let ((w (find-cmd-widget-make-n-widget unit nil)))
          (widget-put w :value-get #'widget-field-value-get)
          w)))))

(defun find-cmd-widget-make-time-days-widget (name hint)
  (find-cmd-widget-make-signed-n-widget name hint "days" '<))

(defun find-cmd-widget-make-time-minutes-widget (name hint)
  (find-cmd-widget-make-signed-n-widget name hint "minutes" '<))

(defun find-cmd-widget-make-links-widget (name hint)
  (find-cmd-widget-make-signed-n-widget name hint "links" '=))

(defun find-cmd-widget-make-type-widget (name hint)
  (find-cmd-widget--warn-arg-not-implemented hint)
  (cl-flet ((vget (s) (lambda (_w) (list (list s)))))
    `(checklist :tag ,(symbol-name name)
                :format "file type:\n%v"
                :value-get ;; validate here since :validate it tricky
                ,(lambda (w)
                   (let ((v (widget-checklist-value-get w)))
                     (when (and (eq name 'type)
                                (eq 'L (car find-cmd-widget-current-global-option))
                                (seq-some (lambda (x) (string= (car x) "l")) v))
                       (push "Global option \"-L\"- and -type l (\"symbolic link\") used together"
                             find-cmd-widget-warnings))
                     `(,name ,(mapconcat #'car v ","))))
                :args ((choice-item :tag "regular file"                   :value-inline ,(vget "f"))
                       (choice-item :tag "directory"                      :value-inline ,(vget "d"))
                       (choice-item :tag "symbolic link"                  :value-inline ,(vget "l"))
                       (choice-item :tag "block (buffered) special"       :value-inline ,(vget "b"))
                       (choice-item :tag "character (unbuffered) special" :value-inline ,(vget "c"))
                       (choice-item :tag "named pipe (FIFO)"              :value-inline ,(vget "p"))
                       (choice-item :tag "socked"                         :value-inline ,(vget "s"))
                       (choice-item :tag "door (Solaris)"                 :value-inline ,(vget "D"))))))

(defun find-cmd-widget-make-combiner-widget (combiner hint)
  (find-cmd-widget--warn-arg-not-implemented hint)
  (let ((sname (symbol-name combiner)))
    `(find-expr :tag ,sname
                :format ,(concat sname "\n%v")
                :value-inline ,(lambda (w) `((,combiner ,@(widget-child-value-get w)))))))

(define-widget 'find-cmd-widget-file-widget 'item
  "Doc..."
  :action (lambda (w &rest _)
            (widget-value-set w (funcall (widget-get w :read-file-name-fun) w))
            (widget-apply w :notify w))
  :read-file-name-fun (lambda (w) (read-file-name "Dir: " nil (widget-value-value-get w)))
  :value "")

;; FIXME: Doesn't work
(defun find-cmd-widget-make-xnewer-widget (name hint)
  `(find-cmd-widget-file-widget
    :value-get ,(lambda (w) `(,name ,(expand-file-name (widget-value-value-get w))))
    :tag ,(symbol-name name)
    :format ,(concat "%[file%]: %v" (when hint (format "          [%s]" hint)) "\n")))

(define-widget 'find-expr 'lazy
  "Doc..."
  :format "...:\n%v"
  :type `(editable-list
          ;; :entry-format "%i\n%d %v" ;; that breaks intendation; %n not
          ;; implemented in `widget-editable-list-entry-create' :-(
          :args
          ((menu-choice
            :tag "expr"
            :explicit-choice ,(find-cmd-widget--get-arg (assoc 'iname find-cmd-widget-find-arg-alist))
            :value "*"
            :action find-cmd-widget-menu-choice-action
            :args ,(mapcar #'find-cmd-widget--get-arg find-cmd-widget-find-arg-alist)))))

(define-widget 'find-cmd-widget-global-option 'menu-choice
  "Doc..."
  :format "global options: %v\n"
  :action 'find-cmd-widget-menu-choice-action
  :notify (lambda (w &rest args)
            (setq find-cmd-widget-current-global-option (widget-value w))
            (apply #'widget-default-notify w args))
  :args (cons `(choice-item :tag "none"
                            :value-get ,(lambda (_w) nil))
              (mapcar #'find-cmd-widget--get-arg
                      find-cmd-widget-find-global-option-arglist)))

(define-widget 'find 'group
  "Doc..."
  :format "%v"
  :value-get (lambda (w)
               (setq find-cmd-widget-warnings '())
               (widget-editable-list-value-get w))
  :notify (pcase-lambda ((app widget-value v) &rest _)
            (widget-value-set find-cmd-widget-preview-widget-1
                              (string-trim-right
                               (pp-to-string
                                (find-cmd-widget-create-find-cmd v))))
            (widget-value-set find-cmd-widget-preview-widget-2
                              (find-cmd-widget-create-find-call v))
            (widget-value-set find-cmd-widget-warning-widget
                              (if find-cmd-widget-warnings
                                  (mapconcat
                                   (apply-partially #'concat "Warning: ")
                                   find-cmd-widget-warnings "\n")
                                "")))
  :args
  `((editable-list
     :format "paths:\n%v%i\n"
     :args
     ((find-cmd-widget-file-widget
       :format "find %[path%]: %v\n"
       ;; initialize VALUE dynamically:
       :convert-widget
       ,(lambda (widget)
          (widget-put widget :value default-directory)
          (widget-put widget :args nil)
          widget)
       :read-file-name-fun ,(lambda (w) (read-directory-name "Dir: " (widget-value w))))))
    (find-cmd-widget-global-option)
    (find-expr :format "expressions:\n%v")))


;;;; Main commands

;;;###autoload
(defun find-cmd-widget ()
  "Doc..."
  (interactive)
  (let ((buf (generate-new-buffer "*Widget Find*")))
    (pop-to-buffer buf)
    (kill-all-local-variables)
    (let ((inhibit-read-only t))
      (erase-buffer))
    (remove-overlays)
    (custom--initialize-widget-variables)
    (setq-local find-cmd-widget-main-widget (widget-create 'find))
    (setq-local find-cmd-widget-warning-widget
                (widget-create 'find-cmd-preview
                               :value ""
                               :value-face 'error))
    (let ((arrow ?⇩))
      (when (char-displayable-p arrow)
        (insert (propertize (concat "    " (string arrow)) 'face '(:height 2.0)) "\n"))
      (setq-local find-cmd-widget-preview-widget-1
                  (widget-create 'find-cmd-preview
                                 :value "Here will appear a `find-cmd' preview"))
      (insert "                     ")
      (widget-create 'push-button
                     :tag "Copy!"
                     :help-echo "Copy `find-cmd' call to kill ring"
                     :action (lambda (&rest _)
                               (let ((s (widget-value find-cmd-widget-preview-widget-1)))
                                 (kill-new s)
                                 (message "%s" s))))
      (insert "\n")
      (when (char-displayable-p arrow)
        (insert (propertize (concat "    " (string arrow)) 'face '(:height 2.0)) "\n"))
      (setq-local find-cmd-widget-preview-widget-2
                  (widget-create 'find-cmd-preview :value "\
Here will appear a preview for the constructed find(1) call"))
      (insert "                     ")
      (widget-create 'push-button
                     :tag "Copy!"
                     :help-echo "Copy \"find\" call to kill ring"
                     :action (lambda (&rest _)
                               (let ((s (widget-value find-cmd-widget-preview-widget-2)))
                                 (kill-new s)
                                 (message "%s" s)))))
    (insert "\n\n\n")
    (widget-create 'push-button
                   :tag "find!"
	           :help-echo "Run constructed find command"
                   :action #'find-cmd-widget-find-action)
    (insert "  ")
    (widget-create 'push-button
                   :tag "find-dired!"
	           :help-echo "Run constructed find command as find-dired"
                   :action #'find-cmd-widget-find-dired-action)
    (insert "\n")
    (use-local-map (let ((map (make-sparse-keymap)))
                     (set-keymap-parent map widget-keymap)
                     (define-key map [(control ?c) (control ?c)]
                       #'find-cmd-widget-find-dired-action)
                     map))
    (widget-setup)
    (goto-char (point-min))))

(provide 'find-cmd-widget)
;;; find-cmd-widget.el ends here
